# SPDX-License-Identifier: MIT

# for certs_check_onetime
. "$scripts_dir/certificates.sh"

# replace numfmt...
numfmt_to_bytes() {
	# remove B suffix if any
	local size="${1%[bB]}" unit

	case "${size%[iI]}" in
	# nondigit then digit, two or more non-digits, starting with non-digit
	*[!0-9]*[0-9]|*[!0-9]*[!0-9]*|[!0-9]*)
		error "Unexpected size format $size (expected e.g. 300M)";;
	0*) error "$size cannot start with 0";;
	*[0-9]) echo "$size"; return;;
	esac

	unit="${size##*[0-9]}"
	size="${size%%[!0-9]*}"
	case "$unit" in
	[Kk]) echo "$((size * 1000))";;
	[Kk][Ii]) echo "$((size * 1024))";;
	[Mm]) echo "$((size * 1000000))";;
	[Mm][Ii]) echo "$((size * 1048576))";;
	[Gg]) echo "$((size * 1000000000))";;
	[Gg][Ii]) echo "$((size * 1073741824))";;
	*) error "unsupported unit in $size (only support G/M/K)";;
	esac
}

roundup() {
	local size="$1" divisor="$2" unit="$3"
	local subdigit

	[ "$size" -ge "$divisor" ] || return 1

	subdigit="$((((size % divisor) + (divisor/10-1))/(divisor/10)))"
	size=$((size / divisor))
	if [ "$size" -ge 10 ]; then
		[ "$subdigit" != 0 ] && size=$((size+1))
		subdigit=""
	fi
	echo "$size${subdigit:+.$subdigit}$unit"
}
numfmt_to_human() {
	local size="$1" suffix="$2"
	roundup "$size" 1000000000 "G$suffix" && return
	roundup "$size" 1000000 "M$suffix" && return
	roundup "$size" 1000 "K$suffix" && return
	echo "$size$suffix"
}


installer_is_sd_installer() {
	# quick check that we can use fs on sd card as is
	info "Checking if ${output_file:-$output} can be used safely..."
	if ! btrfsck --readonly --check-data-csum "${output}p1" >/dev/null 2>&1 \
	    || ! fstype=btrfs mount_rootfs "${output}p1" mnt; then
		info "It looks like your $installer_what does not contain an installer image"
		return 1
	fi

	# mnt/etc/init.d/install was moved to mnt/installer.sh in 3.17-at.7 (2023/06)
	# can simplify the check in a bit more.
	if ! { [ -e "mnt/lib/rc/sh/functions-atmark.sh" ] \
	    && { [ -e "mnt/etc/init.d/install" ] || [ -e "mnt/installer.sh" ]; }; }; then
		umount mnt
		info "It looks like your $installer_what does not contain an installer image"
		return 1
	fi
	# min versions:
	# - 3.15.4-at.7 appfs.lzo restore
	# - 3.19.1-at.1 regenerate sshd host keys
	local version min_version="3.19.1-at.1"
	version=$(cat mnt/etc/atmark-release 2>/dev/null)
	if [ -z "$version" ] || ! printf "%s\n" "$min_version" "$version" | sort -Vc 2>/dev/null; then
		umount mnt
		info "It looks like your installer is too old (found '$version', need at least '$min_version')"
		return 1
	fi
	umount mnt

	if [ -z "$noprompt" ]; then
		echo "An installer system is already available on $installer_what. Use it? [Y/n]"
		prompt_yesno y
	else
		info "Using installer image on $installer_what."
	fi
}

installer_download_aac() {
	local zip_approx_size="" image_approx_size=""
	local retry

	if [ -z "$installer_image_url" ]; then
		case "$(cat /etc/hwrevision)" in
		iot-a6e*)
			installer_image_url="https://armadillo.atmark-techno.com/files/downloads/armadillo-iot-a6e/image/baseos-6e-installer-latest.zip"
			zip_approx_size=170M
			image_approx_size=600M
			;;
		a610*|a640*)
			installer_image_url="https://armadillo.atmark-techno.com/files/downloads/armadillo-640/image/baseos-600-installer-latest.zip"
			zip_approx_size=130M
			image_approx_size=420M
			;;
		*)
			# default go g4
			installer_image_url="https://armadillo.atmark-techno.com/files/downloads/armadillo-iot-g4/image/baseos-x2-installer-latest.zip"
			zip_approx_size=150M
			image_approx_size=450M
			;;
		esac
	fi

	# automatic download for lazy people
	local image="${installer_image_url##*/}"
	local domain="${installer_image_url#http*://}"
	domain="${domain%%/*}"
	# default to y only on empty output
	local dl_default=n dl_prompt="y/N"
	if [ -n "$output_file" ] && [ "$(stat -c %b "$output_file")" = 0 ]; then
		dl_default=y
		dl_prompt="Y/n"
	fi
	echo "Download $image image from $domain${zip_approx_size:+ (~$zip_approx_size)} ? [$dl_prompt]"
	[ "$dl_default" = y ] || echo "WARNING: it will overwrite your $installer_what!!"
	prompt_yesno $dl_default || error "Please write an install disk to mmc card manually"

	if [ -n "$image_approx_size" ] \
	    && [ "$(numfmt_to_bytes "$image_approx_size")" -gt "$output_size" ]; then
		[ -n "$output_file" ] \
			&& [ "${output#/dev/loop}" != "$output" ] \
			|| error "$image requires at least ${image_approx_size}B to fit into $output ($(
					numfmt_to_human "$output_size" B
				))"
		info "Growing $output_file to fit $image ($image_approx_size)"
		losetup -d "$output" \
			|| error "Could not release $output"
		truncate -s "$image_approx_size" "$output_file" \
			|| error "Could not resize $output_file"
		retry=0
		while true; do
			output=$(losetup -f) || error "Could not find a free loop device"
			losetup -P "$output" "$output_file" && break
			retry=$((retry + 1))
			[ $retry -lt 10 ] || error "Could not setup loop device"
			sleep 1
		done
	fi

	info "Downloading and extracting image to $installer_what..."

	# try to get md5 first to fail early
	curl -sSf -O "$installer_image_url.md5" \
		|| error "Could not download image checksum"
	sed -i -e 's/baseos.*/-/' "$image.md5"

	# We're writing the image through many, many pipes as follow:
	# curl | tee (zip_tee) | md5sum (zip_md5) > /dev/null
	#                      | unzip -l (zip_l) | awk (zip_sz) > zip_sz
	#                      | unzip -p (zip_p) | tee (image_tee) | xxh (image_xxh) > xxh
	#                                                           | wc -c (image_wc) > sz
	#                                                           | dd (image_dd) > $output
	# (name) is used for pid, as well as *input* pipe (note curl has no input);
	# command is built in the same order except curl coming last.
	local pid_zip_tee pid_zip_md5 pid_zip_l pid_zip_sz pid_zip_p
	local pid_image_tee pid_image_xxh pid_image_wc pid_image_dd
	mkfifo pipe_zip_tee pipe_zip_md5 pipe_zip_l pipe_zip_sz pipe_zip_p
	mkfifo pipe_image_tee pipe_image_xxh pipe_image_wc pipe_image_dd

	tee pipe_zip_l pipe_zip_md5 > pipe_zip_p < pipe_zip_tee &
	pid_zip_tee=$!

	md5sum -c "$image.md5" < pipe_zip_md5 >/dev/null 2>&1 &
	pid_zip_md5=$!

	unzip -l pipe_zip_l > pipe_zip_sz &
	pid_zip_l=$!

	awk '/.img$/ { print $1, $NF }' < pipe_zip_sz > zip_sz &
	pid_zip_sz=$!

	unzip -p pipe_zip_p '*.img' > pipe_image_tee &
	pid_zip_p=$!

	tee pipe_image_wc pipe_image_xxh > pipe_image_dd < pipe_image_tee &
	pid_image_tee=$!

	xxhsum < pipe_image_xxh > xxh &
	pid_image_xxh=$!

	wc -c < pipe_image_wc > image_sz &
	pid_image_wc=$!

	dd if=pipe_image_dd of="$output" bs=1M oflag=direct \
		     iflag=fullblock status=none &
	pid_image_dd=$!

	# let's rock, and check.
	curl -sSf "$installer_image_url" > pipe_zip_tee \
		|| error "Download of $installer_image_url failed"

	wait "$pid_zip_tee" || error "Could not deduplicate zip stream (zip tee)"
	wait "$pid_zip_md5" || error "Downloaded file doesn't match checksum (zip md5)"
	wait "$pid_zip_l" || error "Could not list archive content (zip -l)"
	wait "$pid_zip_sz" || error "Could not extract image size from zip (zip_sz)"
	wait "$pid_zip_p" || error "Could not extract image from archive (zip -p)"
	wait "$pid_image_tee" || error "Could not deduplicate image stream (image tee)"
	wait "$pid_image_xxh" || error "Could not compute image checksum (image_xxh)"
	wait "$pid_image_wc" || error "Could not compute image size (wc -c)"
	wait "$pid_image_dd" || error "Could not write image to $output (dd)"

	# check size in zip and size written match
	# zip_sz contains original image name in second word
	local zip_sz image_name image_sz
	zip_sz=$(cat zip_sz) && [ -n "$zip_sz" ] \
		|| error "Could not read installer size from zip"
	image_name="${zip_sz#* }"
	zip_sz="${zip_sz%% *}"
	image_sz=$(cat image_sz) && [ -n "$image_sz" ] \
		|| error "Could not read installer size"
	[ "$image_sz" = "$zip_sz" ] \
		|| error "Installer size did not match (expected $zip_sz, got $image_sz)"

	info "Finished writing $image_name, verifying written content..."

	dd if="$output" bs=1M count="$image_sz" \
			 iflag=direct,count_bytes status=none \
		| xxhsum -c xxh >/dev/null \
		|| error "Written installer checksum does not match!"

	rm -f pipe_zip_tee pipe_zip_md5 pipe_zip_l pipe_zip_sz pipe_zip_p \
		 pipe_image_tee pipe_image_xxh pipe_image_wc pipe_image_dd \
		 zip_sz image_sz xxh
}

installer_resize_and_mount_sd_root() {
	# installer has no other partition, just make it as big as possible
	# (we need bigger for appfs dump)
	local part_start part_end disk_end
	local partition_size leftover_size
	local user_partition=""

	# get current main partition start, end and disk size.
	part_start=$(sgdisk -p "$output" | awk -v disk_end=0 '
			/^Disk \/dev/ { disk_end=$3 }
			/^ *1 .*rootfs_0/ {
				part_start=$2;
				if (!part_end) {
					part_end=$3;
				}
			}
			# 2 partitions: mark end as -1
			/^ *2/ { part_end=-1 }
			# fail if > 2 partitions
			/^ *(1[0-9]|[03-9])/ { exit(1) }
			# fail if sector size is not 512
			/Sector size/ && $4 != "512/512" { exit(2); }
			END {
				if (part_start) {
					print part_start, part_end, disk_end
				}
			}')
	case "$?" in
	0) ;;
	1) error "sd installer should only have one or two partitions";;
	2) error "$installer_what card should have 512 byte block size";;
	*) error "unknown error while trying to get $installer_what partitions";;
	esac

	[ -n "$part_start" ] || error "Could not find installer main partition"
	# split variable by space in ugly sh...
	disk_end="${part_start##* }"
	part_start="${part_start% *}"
	part_end="${part_start##* }"
	part_start="${part_start% *}"
	[ "$disk_end" != 0 ] || error "Could not parse disk total size"

	# Round down disk_end to a multiple of 1MB for later compuations
	# This should be noop.
	disk_end=$(((disk_end/2048)*2048))
	# sizes in MB. disk size is leftover disk size, rounded down
	leftover_size=$(((disk_end - part_end) / 2048))

	# already resized: mount and stop here if two partitions (end=-1) or
	# trailing space is < 16MB
	if [ "$part_end" = "-1" ] || [ "$leftover_size" -lt 16 ]; then
		fstype=btrfs mount_rootfs "${output}p1" mnt \
			|| error "Could not mount installer partition"
		return
	fi

	echo "Would you like to create a windows partition?"
	echo "That partition would only be used for customization script at the end of"
	echo "install, leave at 0 to skip creating it."
	printf "%s${noprompt:+0\n}" "Custom partition size (MB, [0] or 16 - $leftover_size): "
	while true; do
		if [ -z "$noprompt" ]; then
			read -r partition_size
		fi
		[ -n "$partition_size" ] || partition_size=0
		[ "$partition_size" = 0 ] && break
		case "$partition_size" in
		*[!0-9]*)
			echo "Please enter digits only"
			continue
			;;
		esac
		if [ "$partition_size" -gt "$leftover_size" ]; then
			echo "Please enter a size < $leftover_size"
			continue
		fi
		if [ "$partition_size" -lt "16" ]; then
			echo "Please enter a size > 16"
			continue
		fi
		break
	done

	# use set / $@ for sgdisk user partition arguments
	if [ "$partition_size" = "0" ]; then
		# no user partition: use all disk
		part_end=""
		set --
	else
		local new_end
		# size in sectors. Last sector is inclusive, so
		# must be one less than a multiple of 2048 for alignment
		partition_size=$((partition_size * 2048))
		new_end=$((disk_end - partition_size - 1))
		[ "$new_end" -ge "$part_end" ] \
			|| error "Sorry, rounding made main partition smaller than it currently is" \
				"run again with a smaller user partition"
		part_end="$new_end"
		set -- --new "2:$((part_end + 1)):" -c "2:install_data"
		user_partition=1
	fi

	info "Checking and growing installer main partition"

	# resize GPT headers first; any command would work here but
	# we cannot specify a $part_end > previous end size without this
	sgdisk -G "$output" >/dev/null \
		&& sgdisk -d 1 --new "1:$part_start:$part_end" -c 1:rootfs_0 \
			"$@" "$output" >/dev/null \
		|| error "resizing partitions failed"

	if [ -n "$user_partition" ]; then
		local mkfs=exfat
		if ! command -v mkfs.exfat; then
			info "Trying to install mkfs.exfat (exfatprogs) in memory from internet"
			if ! apk add exfatprogs; then
				echo "Could not install exfatprogs, format partition using vfat instead? [Y/n]"
				if prompt_yesno y; then
					mkfs=vfat
				else
					warning "Skipping user partition format. Please format ${output}p2 manually"
					mkfs=skip
					user_partition=""
				fi
			fi
		fi
		case "$mkfs" in
		exfat)
			mkfs.exfat -L INST_DATA "${output}p2";;
		vfat)
			mkfs.vfat -n INST_DATA "${output}p2";;
		esac
	fi

	fstype=btrfs mount_rootfs "${output}p1" mnt \
		|| error "Could not mount installer partition"
	local fstype
	fstype=$(findmnt -nr -o FSTYPE mnt)

	# also copy installer_overrides sample to user partition if set...
	if [ -n "$user_partition" ] && mount -t "$mkfs" "${output}p2" mnt/mnt; then
		if [ -e mnt/installer_overrides.sh.sample ]; then
			cp mnt/installer_overrides.sh.sample mnt/mnt/
		fi
		if [ -e mnt/ip_config.txt.sample ]; then
			cp mnt/ip_config.txt.sample mnt/mnt/
		fi
		umount mnt/mnt
	fi

	case "$fstype" in
	ext4)
		umount mnt \
			|| error "Could not umount installer partition"
		e2fsck -f "${output}p1" \
			|| error "e2fsck failed after growing partition"

		# This requires network for ext4 based installer,
		# but any installer used with this should be btrfs
		command -v resize2fs > /dev/null \
			|| apk add e2fsprogs-extra \
			|| error "Need resize2fs (e2fsprogs-extra) to grow installer"
		resize2fs "${output}p1" \
			|| error "resize2fs failed"

		mount -t ext4 "${output}p1" mnt \
			|| error "Could not mount installer partition after growing"
		;;
	btrfs)
		btrfs filesystem resize max mnt \
			|| error "btrfs filesystem resize failed"
		;;
	*)
		error "Unexpected fstype: $fstype. Only ext4/btrfs supported for installer"
		;;
	esac
}

installer_fix_console() {
	# If the console used isn't the default console
	# also set installer console
	local console
	console=$(grep -oE 'console=\S*' /proc/cmdline)

	# console not found in cmdline
	[ -n "$console" ] || return
	# no default env, cannot fix
	[ -e mnt/boot/uboot_env.d/00_defaults ] || return
	# this is the default value
	grep -qxF "$console" mnt/boot/uboot_env.d/00_defaults && return

	info "Setting console to $console in installer"
	echo "$console" > mnt/boot/uboot_env.d/ZZ_installer \
		|| error "Could not write to mnt/boot/uboot_env.d/ZZ_installer"
	cat mnt/boot/uboot_env.d/* \
		| grep -v "upgrade_available=1" \
		| fw_setenv --config mnt/etc/fw_env.config \
			--script - --defenv /dev/null \
		|| error "Could not set installer uboot env"
}

installer_to_lzo_and_csum() {
	# reads piped data, writes lzo and xxh file
	# build-rootfs's build_image.sh's copy_to_lzo_and_csum
	# almost verbatim (sudos removed)
	local dest="$1"
	local sz="$2"
	shift 2
	local xxh
	local pid_xxh pid_wc pid_tee pid_source

	# assume we can read source multiple times
	mkfifo pipe_tee pipe_xxh pipe_lzop pipe_wc
	xxhsum < pipe_xxh > xxh &
	pid_xxh=$!

	wc -c < pipe_wc > sz &
	pid_wc=$!

	tee pipe_lzop pipe_xxh > pipe_wc < pipe_tee &
	pid_tee=$!

	"$@" > pipe_tee &
	pid_source=$!

	lzop > "mnt/$dest.lzo" < pipe_lzop \
	      || error "Failed compressing or writing lzo file"
	wait "$pid_source" \
		|| error "Source command failed for $dest"
	wait "$pid_xxh" \
		|| error "Computing xxh failed for $dest"
	wait "$pid_wc" \
		|| error "wc failed for $dest"
	wait "$pid_tee" \
		|| error "tee failed for $dest"
	if [ -n "$sz" ]; then
		[ "$sz" = "$(cat sz)" ] \
			|| error "Stream was not of the expected size"
	else
		sz="$(cat sz)" \
			|| error "Could not read size"
		# sanity check
		[ -n "$sz" ] && [ "$sz" != 0 ] \
			|| error "Read size was 0, problem with pipes?"
	fi
	xxh=$(cat xxh) \
		|| error "Could not read xxh"
	xxh=${xxh%% *}
	wait "$pid_wc" \
		|| error "$dest file does not have expected size (expected $sz)"
	echo "$sz $xxh" > "mnt/$dest.xxh" \
		|| error "Could not write $dest checksum"
	rm -f pipe_lzop pipe_xxh pipe_wc pipe_tee xxh sz
	local xxhcheck
	xxhcheck="$(lzop -d < "mnt/$dest.lzo" | xxhsum)"
	xxhcheck="${xxhcheck%% *}"
	[ "$xxh" = "$xxhcheck" ] \
		|| error "Sha we just wrote does not match (expected $xxh got $xxhcheck"
}

installer_dump_bootloader() {
	# copy of mkswu's pre_boot.sh copy_boot_imxboot,
	# with extra checks (fw_env.config, encryption)
	local env_offset sz_mb

	# skip if no fw_env.config
	if ! [ -e /etc/fw_env.config ]; then
		return
	fi

	# We copy until env start (assume nothing else than boot image before)
	# Env will be cleared in post_boot
	env_offset=$(awk '/^[^#]/ && $2 > 0 {
			if (!start || $2 < start)
				start = $2;
		}
		END {
			if (!start) exit(1);
			printf("%d\n", start);
		}
		' < /etc/fw_env.config) \
		|| error "Could not get boot env location"

	# use temporary file to check uboot is not encrypted, it's small anyway.
	dd if="${rootdev}boot$ab" of=boot bs=1M count="$env_offset" \
			iflag=count_bytes status=none \
		|| error "Could not read boot image from ${rootdev}boot$ab"

	if ! strings boot | grep -qE 'U-Boot 20[0-9]{2}.[0-1][0-9]'; then
		warning "Could not recognize boot image (encrypted?), keeping installer version"
		return
	fi

	info "Copying boot image"

	# round up to next MB (clear env on target)
	sz_mb=$(((env_offset - 1) / 1024 / 1024 + 1))
	truncate -s "${sz_mb}M" boot || error "Could not resize boot image"
	installer_to_lzo_and_csum boot "$((sz_mb*1024*1024))" \
		cat boot
	rm -f boot
	echo "boot (make-rootfs on $(date))" > mnt/boot.filename \
		|| error "Could not overwrite image filename"
}

check_rootfs_ro() {
	# check there is no writer
	# fun fact: while remount,rw is local to a private mount, remount,ro affects
	# all instances of the fs.
	# the last unshare with a rw mount disappearing also doesn't make the fs back to
	# ro globally (last column of mountinfo)
	# Since we no longer do a block-level copy we can accept a writable fs, but if files
	# get modified something might not be coherent so make the user check.

	local rootopt
	# rootfs is listed as /dev/root (ext4) or $rootpart (/dev/mmcblk2pX) (btrfs),
	# pick the first match.
	if rootopt=$(awk -v root="$rootpart" '
			$10 == root || $10 == "/dev/root" {
				print $NF
				exit(1)
			}' /proc/self/mountinfo); then
		# note on mountinfo: $7 is optional (zero or more optional fields), we want the
		# partition that is two fields after '-'... But that is hard to do, so assume
		# rootfs always has exactly one optional field (shared:x) and is $10 and warn
		# if we did not find it.
		echo "WARNING: Did not find '$rootpart' in /proc/self/mountinfo, skipping rw check"
		echo "(If rootfs changes while we copy it some data may be corrupted)"
		echo "Continue anyway? [y/N]"
		prompt_yesno n || error "Aborted"
		return
	fi
	case ,"$rootopt", in
	*,rw,*) # persist_file or similar was used, needs check.
		;;
	*)      # already ro: remount should be noop and always work.
		mount -o remount,ro /live/rootfs 2>/dev/null && return
		;;
	esac

	# check if still actively rw for more helpful message
	local writers pid pids=""
	writers="$(awk '$6 ~ /(^|,)rw(,|$)/ && $10 == "'"$rootpart"'" { print FILENAME }' /proc/*/mountinfo)"
	for pid in $writers; do
		pid="${pid#/proc/}"
		pid="${pid%/mountinfo}"
		pids="$pids $pid"
	done

	if [ -n "$pids" ]; then
		# Some processes still have a rw fs - they might not be actually using it,
		# but remounting ro could break later writes so we should not do it.
		echo "WARNING: The following processes are possibly writing in rootfs:"
		ps x | awk -v "list=$pids" '
			BEGIN {
				split(list, array);
				for (i in array)
					pids[array[i]]=1;
			}
			($1 in pids)'
	elif ! mount -o remount,ro /live/rootfs >/dev/null 2>&1; then
		# remount didn't work, something we didn't find is still using the fs
		# (possibly holding deleted files)
		echo "WARNING: Could not remount the rootfs read-only, something might still be modifying it."
	else
		# remount,ro worked, all clear.
		return
	fi
	echo "If rootfs changes while we copy it some data may be corrupted, consider rebooting first."
	echo "Continue anyway? [y/N]"
	prompt_yesno n || error "Aborted"
	rootfs_rw=1
}

installer_dump_rootfs() {
	local avail fstype rootfs_rw=""

	info "Copying rootfs"

	avail=$(findmnt -nr --bytes -o AVAIL /live/rootfs)
	[ -n "$avail" ] || error "Could not find rootfs mount"
	[ "$avail" -gt 1048576 ] \
		|| error "The rootfs is full or almost full ($((avail/1024))KB free)" \
			 "Installer requires some free space to work, please free up at least 1MB and restart"

	check_rootfs_ro

	# try to run fsck as appropriate, no point in saving a broken fs
	fstype=$(findmnt -nr -o FSTYPE /live/rootfs)
	case "$fstype" in
	ext4)   # cannot run fsck if fs is still rw for ext4 (journal not clean)
		if [ -z "$rootfs_rw" ]; then
			e2fsck -f -n "$rootpart" >fsck.log 2>&1 \
				|| error "e2fsck failed for rootfs, aborting: $(cat fsck.log)"
		fi;;
	btrfs) btrfsck --readonly --force "$rootpart" >fsck.log 2>&1 \
		|| error "btrfsck failed for rootfs, aborting: $(cat fsck.log)";;
	*) error "Unknown fstype for rootfs, aborting";;
	esac
	rm -f fsck.log

	# Copy into a temporary fs
	truncate -s "$rootpart_size" rootpart

	# XXX share with rollback-clone in later version
	local extlinux=""
	[ -e "/boot/extlinux.conf" ] && extlinux=1
	case "$fstype" in
	ext4)
		mkfs.ext4 -q ${extlinux:+-O "^64bit"} -L "rootfs_0" -F rootpart
		;;
	btrfs)
		mkfs.btrfs -q -L "rootfs_0" -m dup -f rootpart
		;;
	esac || error "Could not create fs for rootfs copy"

	mount_rootfs rootpart rootfs || error "Could not mount rootfs target for copy"

	unshare -m sh -c 'mount --bind /live/rootfs /mnt && cp -a /mnt/. rootfs/' \
                || error "Could not copy existing fs over"

	# installer will fix that anyway, but might as well make a clean rootfs_0
	if [ -e rootfs/etc/fw_env.config ]; then
		sed -i -e "s/boot[01]/boot0/" rootfs/etc/fw_env.config \
	                || error "Could not update target fw_env.config"
	fi
        sed -i -e "s/boot_[01]/boot_0/" rootfs/etc/fstab \
                || error "Could not update target fstab"

	if [ -z "$insecure" ]; then
		# remove ssh host keys and mark for regeneration
		if stat rootfs/etc/ssh/ssh_host_*key* >/dev/null 2>&1; then
			rm -f rootfs/etc/ssh/ssh_host_*key* \
				&& touch rootfs/etc/ssh/ssh_host_keys_installer_regenerate \
				|| error "Could not remove ssh host keys"
		fi
		# close node-red admin interface
		if [ -e rootfs/etc/atmark/containers/node-red.conf ]; then
			sed -i -e '/add_ports 1880:1880/d' \
				rootfs/etc/atmark/containers/node-red.conf
		fi
	fi

	umount rootfs || error "Could not umount rootfs copy"

	installer_to_lzo_and_csum image "$rootpart_size" \
		dd if=rootpart bs=1M status=none

	rm -f rootpart

	echo "rootfs (make-rootfs on $(date))" > mnt/image.filename \
		|| error "Could not overwrite image filename"
}

installer_dump_firmware() {
	local size dev csum

	dev=$(findmnt -nr -o SOURCE /opt/firmware)
	if [ -z "$dev" ]; then
		# only print warning if we are sure there should be one
		# (but we dump if found regardless, we also do not check
		# the value matches what we found...)
		if grep -qE '^HAS_OPT_FIRMWARE' /etc/atmark/baseos.conf 2>/dev/null; then
			warning "/opt/firmware was not mounted, copy skipped"
		fi
		return
	fi
	# we can get the squashfs size through df to not copy too many useless trailing data
	size=$(findmnt -nr --bytes -o SIZE /opt/firmware)
	[ -n "$size" ] || error "Could not get size from /opt/firmware"

	info "Copying /opt/firmware filesystem"
	# this is already squashfs-ified so doesn't use to_lzo_and_csum...
	# do in two steps like build_image, it is small enough.
	dd if="$dev" of=mnt/firm.squashfs bs=1M count="$size" iflag=count_bytes status=none \
		|| error "Could not copy firmware to $installer_what"
	csum=$(xxhsum -q mnt/firm.squashfs) \
		|| error "Could not compute checksum for firmware partition"
	echo "${csum%% *}" > mnt/firm.squashfs.xxh \
		|| error "Could not write firmware checksum"
}

installer_btrfs_del_snap() {
	local snap
	for snap; do
		! [ -d "app/snapshots/$snap" ] \
			|| btrfs -q subvolume del "app/snapshots/$snap" \
			|| error "Could not delete snaphsots/$snap subvolume in appfs"
	done
}

installer_btrfs_snap() {
	local src="$1"
	local dst="$2"

	btrfs -q subvolume snapshot -r "$src" "app/snapshots/$dst" \
		|| error "Could not create snapshot for $src"
	snapshots="$snapshots app/snapshots/$dst"
}

installer_dump_appfs() {
	local appdev snapshots=""

	info "Copying appfs"

	get_appdev

	mount -t btrfs "$appdev" app \
		|| error "Could not mount appfs root"

	installer_btrfs_del_snap volumes boot_volumes \
		containers_storage boot_containers_storage

	[ -d app/snapshots ] \
		|| mkdir app/snapshots \
		|| error "Could not create snapshot dir in appfs"

	installer_btrfs_snap /var/app/volumes volumes
	installer_btrfs_snap /var/app/rollback/volumes boot_volumes
	installer_btrfs_snap /var/lib/containers/storage_readonly boot_containers_storage
	is_mountpoint /var/lib/containers/storage \
		&& installer_btrfs_snap /var/lib/containers/storage containers_storage

	# btrfs send outputs both status messages and errors to stderr,
	# so filter manually to keep only real errors on stderr
	mkfifo pipe_btrfs_send
	awk '{ if ($0 ~ /^At subvol/) { print } else { print >"/dev/stderr" } }' < pipe_btrfs_send &
	local awk_pid=$!
	# shellcheck disable=SC2086 ## snapshots is split on purpose
	installer_to_lzo_and_csum appfs "" \
			btrfs send $snapshots > pipe_btrfs_send 2>&1
	wait "$awk_pid" # ignore awk failure
	rm -f pipe_btrfs_send
}

get_file_holes_size() {
	local file="$1"
	# a sparse file has two "sizes":
	# - the apparent size as reported through e.g. ls
	# - the allocated size as reported through e.g. du
	# This computes the difference between the two, to check how much could
	# still be allocated in this file.
	local apparent allocated blocksize
	apparent=$(stat -c "%s %b %B" "$file" 2>/dev/null) || return
	# break the three words by space...
	blocksize=${apparent##* }
	apparent=${apparent% *}
	allocated=${apparent#* }
	apparent=${apparent% *}

	# allocated is number of blocks * block sizes
	allocated=$((allocated * blocksize))

	# bail out if allocated is bigger than apparent size...
	[ "$apparent" -gt "$allocated" ] || return

	echo $((apparent - allocated))
}

file_is_in_appfs() {
	local file="$1"
	[ "$(findmnt -nr -o UUID --target "$file")" = "$(findmnt -nr -o UUID /var/tmp)" ]
}


installer_checks() {
	local freesize holesize reqsize=0

	[ -b "$output" ] || error "Didn't set up loop device properly? Please report this error."
	if ! partprobe "$output"; then
		error "$output seems used (partprobe could not reload partitions), cleanup any existing mount first"
	fi
	if [ -n "$output_file" ]; then
		# image file: check if there's a loop device for it
		if losetup -a | grep -v "$output:" | grep -qw "$output_file"; then
			error "$output_file found in losetup -a output: cleanup any existing mount first"
		fi
		# Check there is enough space to write image:
		# - if the output file is in appfs, we need to add up the size for later check
		# - otherwise just check now.
		if holesize=$(get_file_holes_size "$output_file"); then
			if file_is_in_appfs "$output_file"; then
				reqsize="$holesize"
			else
				freesize=$(findmnt -nr --bytes -o AVAIL --target "$output_file" 2>/dev/null)
				if [ -n "$freesize" ] && [ "$freesize" -lt "$holesize" ]; then
					echo "WARNING: There might not be enough space to write $output_file"
					echo "WARNING: (wanted $(
							numfmt_to_human "$holesize" B
						) more, got $(
							numfmt_to_human "$freesize" B
						))"
					echo "Continue building image anyway? [y/N]"
					prompt_yesno n || error "Aborting at user request"
				fi
			fi
		fi
	fi
	# fail building installer unless magic variable is set
	[ -n "$INSTALLER_ALLOW_ONETIME_CERT$insecure" ] \
		|| ! certs_has_onetime \
		|| error "Refusing to build installer image with public onetime swupdate certificate installed" \
			"Please install initial_setup.swu (from mkswu --init) first"

	freesize=$(findmnt -nr --bytes -o AVAIL --target /var/tmp 2>/dev/null)
	reqsize=$((reqsize+rootpart_size))
	if [ -n "$freesize" ] && [ "$freesize" -lt "$reqsize" ]; then
		echo "WARNING: You could need up to $(
				numfmt_to_human "$reqsize" B
			) free in /var/tmp to create installer"
		echo "WARNING: Please free some space before running again (currently $(
				numfmt_to_human "$freesize" B
			) available)"
		echo "(You can run \`abos-ctrl rollback-clone\` to free up rollback data if any)"
		echo "Continue building image anyway? [y/N]"
		prompt_yesno n || error "Aborting at user request"
	fi

	output_size=$(blockdev --getsize64 "$output" 2>/dev/null) \
		|| error "You need a SD card plugged in to build installer: nothing found at $output"
	if [ "$output_size" -lt "$used_size" ]; then
		echo "WARNING: Your $installer_what ($(
				numfmt_to_human "$output_size" B
			)) is smaller than estimated max required size ($(
				numfmt_to_human "$used_size" B
			))"
		echo "Continue anyway? [y/N]"
		prompt_yesno n || error "Aborting at user request"
	fi
	if is_mountpoint /var/lib/containers/storage; then
		echo "WARNING: using emmc for podman-storage is a development option and not supported for production"
		echo "WARNING: consider switching back to tmpfs with 'abos-ctrl podman-storage --tmpfs' first"
		echo "Continue building image? [y/N]"
		prompt_yesno n || error "Aborting at user request"
	fi
}

installer_cleanup() {
	trap - EXIT INT QUIT TERM
	[ "${output#/dev/loop}" != "$output" ] && losetup -d "$output"
	[ -n "$tmpdir" ] || return

	if [ "$PWD" = "$tmpdir" ]; then
		is_mountpoint "rootfs" && umount "rootfs"
		is_mountpoint "mnt" && umount "mnt"
		if is_mountpoint "app"; then
			installer_btrfs_del_snap volumes boot_volumes \
				containers_storage boot_containers_storage
			rmdir app/snapshots >/dev/null 2>&1
			umount "app"
		fi
	fi
	find "$tmpdir" -xdev -delete 2>/dev/null
}

installer_trap() {
	info "Terminating, cleaning up..."

	installer_cleanup
	exit 1
}

installer_fstrim() {
	# trim data if we think we can: output is a real device or
	# a filesystem that supports punch hole
	# (any other filesystem would display scary errors in dmesg)
	if [ -n "$output_file" ]; then
		case "$(findmnt -nr -o FSTYPE --target "$output_file")" in
		ext4|xfs|btrfs) ;;
		*) return;;
		esac
	fi

	fstrim "$mountpoint"
}

installer_shrink_install_disk() {
	local mountpoint=mnt
	local part_start part_end
	local new_size new_end

	part_start=$(sgdisk -p "$output" 2>/dev/null | awk '
		/^ *1/ { part_start=$2; part_end=$3 }
		/^ *(1[0-9]|[2-9])/ { exit(1) }
		END { print part_start, part_end }
	') || return 0
	part_end="${part_start##* }"
	part_start="${part_start% *}"

	# shrink the btrfs volume
	info "Trying to shrink the installer partition..."
	while btrfs filesystem resize -200M "$mountpoint" > /dev/null 2>&1; do
		true
	done

	# trim after we're done with resize
	installer_fstrim

	new_size="$(btrfs filesystem show --raw "$mountpoint" | \
		awk '/devid/ {print $4}')"
	# round up to next sector if required (just in case)
	new_end=$((part_start + (new_size + 511) / 512))
	if [ "$new_end" -ge "$part_end" ]; then
		info "Installer partition is not shrinkable"
		return
	fi

	info "Shrinking the installer partition..."
	if ! umount "$mountpoint"; then
		info "Sleeping a while and retrying..."
		sleep 3
		if ! umount "$mountpoint"; then
			warning "Could not unmount the $installer_what to shrink the partition." \
				"Shrink manually if required."
			# don't leave partition half-resize behind
			btrfs filesystem resize max "$mountpoint" >/dev/null 2>&1
			return 0
		fi
	fi

	# resize the partition
	sgdisk -d 1 --new "1:$part_start:$new_end" -c 1:rootfs_0 \
		"$output" >/dev/null \
		|| error "resizing partition failed"

	# if the output is image file, shrink that file
	if [ -n "$output_file" ]; then
		# interact with file from here on
		losetup -d "$output"
		output="$output_file"

		local gpt_table_offset filesize
		gpt_table_offset=$(sgdisk -p "$output" 2>/dev/null \
			| awk '/Main partition table begins/ { print $NF + 1 }')
		[ -n "$gpt_table_offset" ] || error "Could not get gpt partition table offset"
		filesize=$((new_end * 512 + gpt_table_offset * 512))
		truncate -s "$filesize" "$output_file" \
			|| error "resizing image file failed"
		# fix gpt header. This is noisy so always silence output, rerun on error
		sgdisk -e "$output" >/dev/null 2>&1 \
			|| error "fixing GPT header failed: $(sgdisk -e "$output" 2>&1)"
	fi
}

installer_make_archive_with_md5() {
	local gzip_md5 image_file

	# if the output isn't image file, do nothing
	[ -n "$output_file" ] || return

	info "Computing install disk image's checksum..."
	image_file="$(basename "$output_file")"
	(
		cd "$(dirname "$output_file")" \
			|| error "Could not enter the image file directory"
		md5sum "$image_file" >"$image_file.md5-in-progress" \
			&& mv "$image_file.md5-in-progress" "$image_file.md5" \
			|| error "Failed to compute the checksum"

		info "Creating the install disk image archive..."
		tar -czf "$image_file.tgz-in-progress" "$image_file" "$image_file.md5" \
			|| error "Failed to create the archive"
		sync "$image_file.tgz-in-progress" \
			&& mv "$image_file.tgz-in-progress" "$image_file.tgz" \
			|| error "Could not rename temporary archive to $image_file.tgz"
	) || exit

	info "Checking archived image's checksum..."
	gzip_md5=$(dd if="$output_file.tgz" bs=1M iflag=direct status=none \
		| tar -zxO "$image_file" | md5sum)
	[ "$(sed -e "s/$image_file/-/g" "$output_file.md5")" = "$gzip_md5" ] \
		|| error "the image checksum does not match!"
}

make_installer_help() {
	echo "Usage: abos-ctrl make-installer [make-installer options]"
	echo
	echo "make-installer options:"
	echo "  --image-url <installer image URL>: installer disk image"
	echo "  --output <image file path>: installer disk image file path"
	echo "  --image-size <image file size>: specify image file size"
	echo "    (in MiB unless unit is specified)"
	echo "  --make-archive: make archive of image file and md5 checksum"
	echo "  --insecure: leave sensitive files in image"
	echo
	echo "This command writes the install disk content to SD card "
	echo "by default. Use --output to write an image file."
	echo "If SD card or the image file that specified with --output "
	echo "has no valid installer content, download the latest installler "
	echo "disk image from atmark-techno.com for this device. Use "
	echo "--image-url to specify a custom URL."
	echo "Both --image-size and --make-archive can only be used with "
	echo "--output."
	echo "Examples:"
	echo "  abos-ctrl make-installer"
	echo "  abos-ctrl make-installer --output /mnt/installer.img"
	echo "  abos-ctrl make-installer --output /mnt/installer.img \\"
	echo "    --image-size 600MiB"
	echo "  abos-ctrl make-installer --output /mnt/installer.img \\"
	echo "    --make-archive"
	echo "  abos-ctrl make-installer --output /mnt/installer.img \\"
	echo "    --image-size 700 --make-archive"
}

ctrl_make_installer() {
	local tmpdir="" used_size existing_size=0
	local output_size output_file="" retry
	local output="" image_size="" make_archive=""
	local installer_image_url="" insecure=""
	local installer_what="SD card"
	local noprompt="" rootpart_size

	while [ "$#" -gt 0 ]; do
		case "$1" in
		"--image-url")
			[ "$#" -ge 2 ] || error "$arg requires an argument";
			installer_image_url="$2"
			shift 2
			;;
		"--output")
			[ "$#" -ge 2 ] || error "$arg requires an argument";
			output="$2"
			shift 2
			;;
		"--image-size")
			[ -n "$output" ] || error "$arg is only allowed with --output"
			[ "$#" -ge 2 ] || error "$arg requires an argument";
			image_size="$2"
			case "$image_size" in
			*[0-9])
				# assume MiB
				image_size=$((image_size * 1048576))
				;;
			*)
				image_size="$(numfmt_to_bytes "$image_size")" \
					|| error "$2 is not a valid size"
				image_size="${image_size%B}"
				;;
			esac
			# mkfs.btrfs will fail below this size
			[ "$image_size" -ge $((109*1048576)) ] || error "$2 must be at least 109MiB"
			shift 2
			;;
		"--make-archive")
			[ -n "$output" ] || error "$arg is only allowed with --output"
			make_archive="yes"
			shift
			;;
		"--noprompt")
			noprompt=1
			shift
			;;
		"--insecure")
			insecure=1
			shift
			;;
		*)
			warning "Invalid option $1"
			make_installer_help >&2
			exit 1
			;;
		esac
	done
	command -v xxhsum > /dev/null \
		|| apk add xxhash \
		|| error "xxhsum command is required and installation from network failed"

	dev_from_current
	rootpart_size=$(blockdev --getsize64 "$rootpart") \
		|| error "Could not get size from $rootpart"

	used_size=$(findmnt -nr --bytes -o USED /var/tmp) \
		|| error "Could not get appfs used size"
	# add 300MB for installer rootfs + rootfs + firm as worst estimate
	used_size=$((used_size + 300*1024*1024))

	trap installer_trap EXIT INT QUIT TERM

	if [ -z "$output" ]; then
		for output in /dev/mmcblk[0-9]; do
			# try likely paths without boot partition
			[ -e "$output" ] && ! [ -e "${output}boot0" ] && break
		done
		[ -e "$output" ] && ! [ -e "${output}boot0" ] \
			|| error "Could not find sd card"
	elif ! [ -b "$output" ]; then
		if [ -z "$image_size" ]; then
			image_size="$used_size"
		fi
		# We don't want user to use a partially written installer, so
		# work in a "-partial" file while we operate.
		# We can reuse such a partial file though.
		if [ -e "$output" ]; then
			mv "$output" "$output-in-progress" \
				|| error "Could not rename $output to $output-in-progress"
		fi
		output="$output-in-progress"
		# note what files to cleanup
		if [ -e "$output" ]; then
			existing_size=$(stat -c %s "$output")
		fi
		if [ "$existing_size" -lt "$image_size" ]; then
			truncate -s "$image_size" "$output" \
				|| error "Could not set $output size"
		elif [ "$existing_size" -gt "$image_size" ]; then
			info "$output ($(
					numfmt_to_human "$existing_size" B
				)) was bigger than $(
					numfmt_to_human "$image_size" B
				) and was not truncated."
		fi
		output_file=$(realpath "$output")
		retry=0
		while true; do
			output=$(losetup -f) || error "Could not find a free loop device"
			losetup -P "$output" "$output_file" && break
			retry=$((retry + 1))
			[ $retry -lt 10 ] || error "Could not setup loop device"
			sleep 1
		done
		installer_what="image file"
	fi

	installer_checks

	tmpdir=$(mktemp -d "/var/tmp/make_installer.XXXXXXX") \
		|| error "Could not create tmpdir"
	cd "$tmpdir" || error "Could not enter tmpdir"
	mkdir mnt app rootfs || error "Could not create dirs in tmpdir"

	btrfs device scan --forget
	installer_is_sd_installer || installer_download_aac
	installer_resize_and_mount_sd_root
	installer_fix_console

	# remove any previous installer_swus
	rm -rf mnt/installer_swus

	installer_dump_bootloader
	installer_dump_rootfs
	installer_dump_firmware
	installer_dump_appfs
	installer_shrink_install_disk

	info "Cleaning up and syncing changes to disk..."

	trap - EXIT INT QUIT TERM
	installer_cleanup
	sync

	if [ -n "$output_file" ]; then
		output_file="${output_file%-in-progress}"
		sync "$output_file-in-progress" \
			&& mv "$output_file-in-progress" "$output_file" \
			|| error "Could not rename $output_file-in-progress back to $output_file"
	fi
	[ "$make_archive" != "yes" ] || installer_make_archive_with_md5

	info "Installer updated successfully!"
	[ -z "$output_file" ] || ls -lh "${output_file}"*
}
